import random
import string

from django.db import IntegrityError


def new_uid(max_length, prefix=None,
            chars=string.ascii_letters + string.digits):
    max_generation_attempts = 3
    prefix = prefix or ''
    ln = max_length - len(prefix)
    for i in range(1, max_generation_attempts + 1):
        try:
            # If only digits are used and no prefix specified,
            # ensure that first digit will never be 0
            if not prefix and chars in string.digits and '0' in chars:
                pure_uid = random.choice(str(chars).replace('0', ''))
                pure_uid += ''.join(random.choice(chars) for n in range(ln - 1))
            else:
                pure_uid = ''.join(random.choice(chars) for n in range(ln))

            # UIDs starting from '111' and 'aaa' are system-reserved and
            # should never be generated by application.
            if pure_uid.startswith('111') or pure_uid.lower().startswith('aaa'):
                raise IntegrityError(
                    'Can not generate UIDs starting from "111" or "aaa", '
                    'they are system-reserved.'
                )

            return prefix + pure_uid
        except IntegrityError:
            if i >= max_generation_attempts:
                raise



from pkg_resources import get_distribution, DistributionNotFound

from .fields import UIDField  # noqa: F401
from .misc import new_uid  # noqa: F401

try:
    _dist = get_distribution('django-uidfield')
except DistributionNotFound:
    __version__ = 'Please install this project with setup.py'
else:
    __version__ = _dist.version
VERSION = __version__   # synonym



import re

from django.db.transaction import atomic
from django.test import TestCase, TransactionTestCase, override_settings

from .fields import UIDField
from .models import UIDModel

FIELD_RE = re.compile('[a-z]{3}_[A-Za-z0-9]{16}')


class TestModel(UIDModel):
    uid_field = UIDField(prefix='tmp_', max_length=20, unique=True)
    nullable_uid_field = UIDField(
        prefix='nul_',
        max_length=20,
        unique=True,
        null=True,
        blank=True,
    )


class TestModelWithPk(UIDModel):
    id = UIDField(primary_key=True, prefix='pk_', max_length=20)


class TestDbRouterDefault:
    def db_for_read(self, model, **hints):
        return 'default'

    def db_for_write(self, model, **hints):
        return 'default'


class TestDbRouterOther:
    def db_for_read(self, model, **hints):
        return 'other'

    def db_for_write(self, model, **hints):
        return 'other'


class CheckMixin:
    def _check_field_value(self, value):
        self.assertEqual(len(value), 20)
        self.assertNotEquals(FIELD_RE.match(value), None)


class UIDFieldTest(CheckMixin, TestCase):
    """UIDField Wrapper Tests"""

    def test_uid_field_value_generate(self):
        """Test generating an UID value in UIDField"""
        obj = TestModel.objects.create()
        new_obj = TestModel.objects.get(id=obj.id)
        self._check_field_value(new_obj.uid_field)

    def test_nullable_uid_field_value_stays_nullable(self):
        """Test generating an UID value in UIDField"""
        obj = TestModel.objects.create()
        new_obj = TestModel.objects.get(id=obj.id)
        self.assertIsNone(new_obj.nullable_uid_field)

    def test_uid_field_value_regeneration(self):
        """Test regenerating an UID value in UIDField"""
        first_obj = TestModel.objects.create()
        second_obj = TestModel(uid_field=first_obj.uid_field)
        second_obj.save()
        self._check_field_value(second_obj.uid_field)
        self.assertNotEquals(first_obj.uid_field, second_obj.uid_field)

    def test_uidfield_as_pk(self):
        """Test usage of UIDField as primary key"""
        obj = TestModelWithPk.objects.create()
        self.assertTrue(obj.pk.startswith('pk_'))
        TestModelWithPk.objects.create()
        another_copy_of_obj = TestModelWithPk.objects.get(pk=obj.pk)
        self.assertEqual(obj, another_copy_of_obj)


class UIDFieldMultidbTest(CheckMixin, TransactionTestCase):
    databases = {'default', 'other'}

    def test_uid_recreation_with_explicit_routing(self):
        """
        Test regenerating an UID value in UIDField in correct DB
        with explicit routing
        """
        for db in {'default', 'other'}:
            # it is important to start transaction outside the create/save
            # methods to check that it is not broken by incorrectly created
            # inner transactions
            with atomic(using=db):
                first_obj = TestModel.objects.using(db).create()
                second_obj = TestModel(uid_field=first_obj.uid_field)
                second_obj.save(using=db)
                self._check_field_value(second_obj.uid_field)
                self.assertNotEquals(first_obj.uid_field, second_obj.uid_field)

    def test_uid_recreation_with_route_based_routing(self):
        """
        Test regenerating an UID value in UIDField in correct DB.
        Routing is based on DB router.
        """
        for db in {'default', 'other'}:
            # route all queries to the selected DB. the save method should
            # correctly determine correct DB where to open transaction
            router = 'django_uidfield.tests.TestDbRouter{db}'.format(
                db=db.capitalize(),
            )
            with atomic(using=db), override_settings(DATABASE_ROUTERS=[router]):
                first_obj = TestModel.objects.create()
                second_obj = TestModel(uid_field=first_obj.uid_field)
                second_obj.save()
                self._check_field_value(second_obj.uid_field)
                self.assertNotEquals(first_obj.uid_field, second_obj.uid_field)



import string

from django.db import models

from .misc import new_uid


class UIDField(models.CharField):
    prefix = None
    chars = None

    def __init__(self, prefix=None, chars=string.ascii_letters + string.digits,
                 *args, **kwargs):
        self.prefix = prefix
        self.chars = chars
        super(UIDField, self).__init__(*args, **kwargs)

    @property
    def non_db_attrs(self):
        return super().non_db_attrs + ('prefix', 'chars')

    def populate(self, model_instance, force_renew=False):
        uid = getattr(model_instance, self.attname, None)
        if (not uid and not self.null) or force_renew:
            uid = new_uid(self.max_length, prefix=self.prefix, chars=self.chars)
            setattr(model_instance, self.attname, uid)
        return uid

    def pre_save(self, model_instance, add):
        if add:
            return self.populate(model_instance)
        else:
            return getattr(model_instance, self.attname)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.prefix:
            kwargs['prefix'] = self.prefix
        kwargs['chars'] = self.chars
        return name, path, args, kwargs



from setuptools import setup, Command

version = '0.3.4'


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from django.conf import settings
        settings.configure(
            DATABASES={
                'default': {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3',
                },
                'other': {
                    'NAME': ':memory:',
                    'ENGINE': 'django.db.backends.sqlite3',
                },
            },
            INSTALLED_APPS=('django_uidfield', 'django.contrib.contenttypes'),
            DEFAULT_AUTO_FIELD='django.db.models.AutoField',
        )
        from django.core.management import call_command
        import django

        django.setup()

        call_command('test', 'django_uidfield')


setup(
    name='django-uidfield',
    version=version,
    description='django-uidfield is a library which includes class '
                'UIDField for models.',
    long_description=open('README.rst').read(),
    keywords='django model field',
    license='MIT',
    author='ivelum',
    author_email='info@ivelum.com',
    url='https://github.com/ivelum/django-uidfield/',
    install_requires=[
        'django',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Plugins',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=['django_uidfield'],
    include_package_data=True,
    cmdclass={'test': TestCommand},
)



from django.db import models, IntegrityError, router, transaction

from .fields import UIDField


class UIDModel(models.Model):
    max_save_attempts = 3

    class Meta:
        abstract = True

    def uid_fields(self):
        return [f for f in self._meta.fields if isinstance(f, UIDField)]

    def populate_uid_fields(self, force_renew=False, uid_fields=None):
        for field in uid_fields or self.uid_fields():
            field.populate(self, force_renew=force_renew)

    def save(self, *args, **kwargs):
        using = (
            kwargs.get('using') or
            router.db_for_write(self._meta.model, instance=self)
        )
        for save_attempt in range(self.max_save_attempts):
            try:
                with transaction.atomic(using=using):
                    return super(UIDModel, self).save(*args, **kwargs)
            except IntegrityError as e:
                uid_fields = self.uid_fields()
                if not any(field.attname in str(e) for field in uid_fields):
                    raise
                if save_attempt < self.max_save_attempts - 1:
                    self.populate_uid_fields(
                        force_renew=True, uid_fields=uid_fields
                    )
        raise IntegrityError('Failed to save %s with unique UID fields' % self)
